/*
 * This program maintains an arbitrary file of acct/passwd pairs using the
 * Unix/Linux crypt facility.  Lines of the file start with acct:pwd, followed
 * by anything, so it can read /etc/passwd or /etc/shadow, but cannot
 * correctly create entries for such files.  It can add entries (with empty
 * "extra" stuff), or peform updates.  These are safe if the size does not
 * change, otherwise they will overwrite following text.
 *
 *  $0 file cmd [ -f ] account pass
 *    cmd =
 *      add		Add the account to the end of the file.
 *      check		Check the password for the account.
 *      update		Update password.
 *  Generally operations refuse and fail if something looks fishy, but -f
 *  makes them soldier on.
 */

#include <crypt.h>
#include <errno.h>

#include <cstring>
#include <cstdlib>

#include <string>
#include <iostream>
#include <fstream>

using std::string;

/*
 * Command summary and die.
 */
void usage(string pname)
{
	std::cerr << "Usage: " << pname << " file cmd [ -f ] account pass"
		  << std::endl;
	std::cerr << "  cmd in { add, check, update }" << std::endl;
	exit(10);
}

/*
 * Read to the indicated user, if any, or eof.  If not found, return "", and 
 * leave write pointer at end-of-file.  If found, return password field and 
 * leave write pointer at start of password.
 */
string find(std::fstream &file, string user)
{
	// Start over.
	file.seekg(0);

	// Scan user name for match.  Read a character at a time until
	// we find the account starting at the start of line, and ending
	// with a colon.
	char inch;		// Input character.
	user += ":";		// Add : so it matches the one in the file.
	auto up = user.begin();	// Scans the name.
	bool found = false;	// Name found flag.
	while(!found && file.get(inch)) {
		if(inch == '\n') {
			up = user.begin();
		} else if(up != user.end()) {
			if(*up++ == inch) {
				if(inch == ':') found = true;
			} else {
				up = user.end();
			}
		}
	}

	// If not found, we'll want to report that.
	if(!found) {
		file.clear();
		file.seekp(file.tellg());
		return "";
	}

	// If we found the userid, we're ready to read the password.
	string pass;				// Password
	std::streampos ploc = file.tellg();	// Where it is in the file.
	
	// Read it.
	while(file.get(inch)) {
		if(inch == '\n' || inch == ':') break;
		pass += inch;
	}

	// Set the file to write a new password if desired.
	file.seekp(ploc);
	return pass;
}

int main(int argc, char **argv)
{
	// Play with parms.
	bool force = (argc == 6 && string(argv[3]) != "-f");
	if(argc < 5 || argc > 6 || (argc == 6 && !force))
		usage(argv[0]);
	string file = argv[1], cmd = argv[2], acct, pass;
	if(force) {
		acct = argv[4];
		pass = argv[5];
	} else {
		acct = argv[3];
		pass = argv[4];
	}

	// See if the command is real.
	if(!(cmd == "add" || cmd == "check" || cmd == "update"))
		usage(argv[0]);

	// Decide on the open mode.
	std::ios_base::openmode openmodes =
		std::fstream::in|std::fstream::binary;
	if(cmd != "check") openmodes |= std::fstream::out;
	
	// Open the file
	std::fstream pswd(file, openmodes);
	if(!pswd) {
		std::cerr << "Unable to open " << argv[1] << std::endl;
		exit(1);
	}

	// Find the entry.
	string encpass = find(pswd, acct);
	if(encpass == "") {
		if(cmd == "check" || cmd == "update") {
			std::cerr << "Account " << acct << " not found."
				  << std::endl;
			exit(2);
		}
	} else {
		if(cmd == "add" && !force) {
			std::cerr << "Account " << acct << " already exists."
				  << std::endl;
			exit(3);
		}
	}

	// Create the pass area, and fill in the requested phrase.
	struct crypt_data pass_data;
	memset(&pass_data, 0, sizeof pass_data);

	// Fill in the setting area.  This chooses the encryption method.
	// For a check or update, use the same method as the existing entry.
	// For an add, generate a new method using system defaults.
	if(cmd == "check" || cmd == "update") {
		// For a check, the current setting.
		strcpy(pass_data.setting, encpass.c_str());
	} else {
		// For anything that sets, generate a new setting.
		char *cr = crypt_gensalt_rn
			(NULL, 0, NULL, 0,
			 pass_data.setting, sizeof pass_data.setting);
		if(cr == NULL) {
			std::cerr << "Password init failed: "
				  << strerror(errno) << std::endl;
			exit(4);
		}
	}

	// Encrypt the phrase.
	char *ret = crypt_r(pass.c_str(), pass_data.setting, &pass_data);
	if(ret == NULL) {
		std::cerr << "Password init failed: "
			  << strerror(errno) << std::endl;
		exit(5);
	}

	// Okay.  See what to do now.
	if(cmd == "check") {
		if(strcmp(pass_data.output, pass_data.setting) == 0) {
			std::cout << "Passwords match." << std::endl;
		} else {
			std::cout << "Passwords do not match."
				  << std::endl;
			exit(20);
		}
	}
	else if(cmd == "update") {
		if(strlen(pass_data.output) != strlen(pass_data.setting) &&
		   !force) {
			std::cout << "New encrypted password is of a "
				  << "different size." << std::endl;
			exit(6);
		}
		pswd.write(pass_data.output, strlen(pass_data.output));
	} else if(cmd == "add") {
		pswd.seekp(0,std::ios_base::end);
		pswd << acct << ":" << pass_data.output << std::endl;
	}

	exit(0);
}
